/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "ForegroundCutoffMeshComponent.h"
#include "Rendering/ExpCamera.h"
#include "Data/SdkData.h"
#include "Data/SharedTexIds.h"

#include "UObject/ConstructorHelpers.h"
#include "Engine/Engine.h"
#include "Engine/World.h"
#include "GameFramework/PlayerController.h"
#include "Engine/StaticMesh.h"
#include "Materials/MaterialInstanceDynamic.h"

using namespace mixcast::data;

TArray<UForegroundCutoffMeshComponent*> UForegroundCutoffMeshComponent::Active = TArray<UForegroundCutoffMeshComponent*>();

UForegroundCutoffMeshComponent::UForegroundCutoffMeshComponent()
{
	LightingChannels = FLightingChannels();

	bTreatAsBackgroundForOcclusion = true;

	bAffectDynamicIndirectLighting = false;
	bAffectDistanceFieldLighting = false;
	bCastDynamicShadow = false;
	bCastStaticShadow = false;
	bReceivesDecals = false;

	static ConstructorHelpers::FObjectFinder<UStaticMesh> Find_PlaneMeshAsset(TEXT("StaticMesh'/Engine/BasicShapes/Plane'"));
	if (Find_PlaneMeshAsset.Succeeded())
	{
		SetStaticMesh(Find_PlaneMeshAsset.Object);
	}

	static ConstructorHelpers::FObjectFinder<UMaterial> Find_ApplyDepthCutoffMat(TEXT("Material'/MixCast/ApplyDepthCutoff.ApplyDepthCutoff'"));
	if (Find_ApplyDepthCutoffMat.Succeeded())
	{
		ApplyDepthCutoffMat = UMaterialInstanceDynamic::Create(
			Cast<UMaterialInterface>(Find_ApplyDepthCutoffMat.Object),
			this,
			TEXT("ApplyDepthCutoff")
		);
		SetMaterial(0, ApplyDepthCutoffMat);
	}
}

void UForegroundCutoffMeshComponent::BeginPlay()
{
	Super::BeginPlay();

	Active.Add(this);

	const UPluginProjectSettings* projectSettings = SdkData::GetProjectSettings();
	FLinearColor clearColor = projectSettings->TranslucencyMode == EMixCastTranslucencyMode::ForceAdditive
		? FLinearColor::Black
		: projectSettings->BackgroundColor;
	ApplyDepthCutoffMat->SetVectorParameterValue(FName("BackgroundColor"), clearColor);
}
void UForegroundCutoffMeshComponent::AttachToCamera(AExpCamera* camera)
{
	this->AttachTo(camera->GetRootComponent(), NAME_None);
	nearPlane = camera->GetNearClipPlane() + 0.01f;
	FVector planeRotation = FVector(-90, 0, -90);
	this->SetRelativeLocationAndRotation(FVector(nearPlane, 0, 0), FQuat::MakeFromEuler(planeRotation));
	ApplyDepthCutoffMat->SetScalarParameterValue(FName("NearClip"), nearPlane);

	APlayerController* player = GEngine->GetFirstLocalPlayerController(GetWorld());
	if (player)
		player->HiddenActors.Add(GetOwner());

	FgCutoffReceiver = new SharedTextureReceiver(SharedTexIds::Cameras::ForegroundCutoff::Get(camera->Identifier));
	TexChangedDelegate = FgCutoffReceiver->OnTextureReferenceChanged().AddUObject(this, &UForegroundCutoffMeshComponent::HandleTextureChanged);
	HandleTextureChanged();
}

void UForegroundCutoffMeshComponent::EndPlay(const EEndPlayReason::Type endPlayReason)
{
	Super::EndPlay(endPlayReason);

	Active.Remove(this);

	if (FgCutoffReceiver)
	{
		FgCutoffReceiver->OnTextureReferenceChanged().Remove(TexChangedDelegate);
		delete FgCutoffReceiver;
		FgCutoffReceiver = nullptr;
	}
}

void UForegroundCutoffMeshComponent::HandleTextureChanged()
{
	UTexture* tex = FgCutoffReceiver->GetTexture();
	if (tex)
		tex->SRGB = 0;
	ApplyDepthCutoffMat->SetTextureParameterValue(FName("DepthCutoff"), tex);
}

void UForegroundCutoffMeshComponent::Update(VirtualCamera* camInfo)
{
	FgCutoffReceiver->EnqueueUpdate();

	float heightAtDist = 2.0f * nearPlane * 0.01f * tanf(FMath::DegreesToRadians(camInfo->fieldOfView * 0.5f));
	float widthAtDist = heightAtDist * (float)camInfo->renderResolutionWidth / camInfo->renderResolutionHeight;
	this->SetRelativeScale3D(FVector(widthAtDist, heightAtDist, 1));
}
